/*
 * Softcam plugin to VDR (C++)
 *
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 */

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>

#if defined(NAGRA_SUPPORT) | defined(CONAX_SUPPORT)
#include <openssl/bn.h>
#endif

#include <vdr/videodir.h>

#include "common.h"
#include "data.h"
#include "misc.h"
#ifdef NAGRA_SUPPORT
#include "libnagra/nagra.h"
#endif

cLoaders loaders;

// -- cLoader ------------------------------------------------------------------

cLoader::cLoader(const char *Id)
{
  id=strdup(Id);
  modified=false;
  loaders.Register(this);
}

cLoader::~cLoader()
{
  free(id);
}

// -- cLoaders -----------------------------------------------------------------

cLoaders::cLoaders(void)
{
  cacheFile=0;
}

cLoaders::~cLoaders()
{
  free(cacheFile);
}

void cLoaders::Register(cLoader *ld)
{
  Lock();
  Add(ld);
  Unlock();
}

void cLoaders::LoadCache(const char *cfgdir)
{
  Lock();
  const char *name=AddDirectory(cfgdir,CACACHE_FILE);
  cacheFile=strdup(name);
  if(access(name,F_OK)==0) {
    isyslog("loading ca cache from %s",name);
    d(printf("loaders: loading ca cache from %s\n",name))
    FILE *f=fopen(name,"r");
    if(f) {
      char buf[256];
      cLoader *ld=0;
      while(fgets(buf,sizeof(buf),f)) {
        if(buf[0]=='#') continue;
        if(!strncmp(buf,":::",3)) { // new loader section
          ld=FindLoader(stripspace(&buf[3]));
          if(!ld) esyslog("unknown loader section '%s'",&buf[3]);
          }
        else if(!strncmp(buf,"ECM:",4) || !strncmp(buf,"KEY:",4)) {  // compatibility, old format
          buf[3]=0; ld=FindLoader(buf);
          if(ld && ld->ParseLine(&buf[4],true)) ld->Modified(); // save the new format right away
          ld=0;
          }
        else if(ld) {
          if(!ld->ParseLine(buf,true)) {
            d(printf("loaders: '%s' failed on line '%s'\n",ld->Id(),buf))
            }
          }
        }
      fclose(f);
      }
    else LOG_ERROR_STR(name);
    }
  Unlock();
}

void cLoaders::SaveCache(void)
{
  Lock();
  if(cacheFile && IsModified()) {
    cSafeFile f(cacheFile);
    if(f.Open()) {
      fprintf(f,"## This is a generated file. DO NOT EDIT!!\n"
                "## This file will be OVERWRITTEN WITHOUT WARNING!!\n");

      cLoader *ld=First();
      while(ld) {
        fprintf(f,":::%s\n",ld->Id());
        if(!ld->Save(f)) break;
        ld=Next(ld);
        }
      f.Close();
      d(printf("loaders: saved cache to file\n"))
      }
    }
  Unlock();
}

bool cLoaders::IsModified(void)
{
  bool res=false;
  Lock();
  cLoader *ld=First();
  while(ld) {
    if(ld->IsModified()) { 
      res=true; break;
      }
    ld=Next(ld);
    }
  Unlock();
  return res;
}

cLoader *cLoaders::FindLoader(const char *id)
{
  Lock();
  cLoader *ld=First();
  while(ld) {
    if(!strcmp(id,ld->Id())) break;
    ld=Next(ld);
    }
  Unlock();
  return ld;
}

// -- cECMInfo -----------------------------------------------------------------

cECMInfo::cECMInfo(void)
{
  name=0;
}

cECMInfo::cECMInfo(cECMInfo *e)
{
  name=strdup(e->name);
  system=e->system;
  ecm_pid=e->ecm_pid;
  id=e->id;
}

cECMInfo::cECMInfo(char *Name, int System, int Pid, int Id)
{
  name=strdup(Name);
  system=System;
  ecm_pid=Pid;
  id=Id;
}

cECMInfo::~cECMInfo()
{
  free(name);
}

// -- cEcmData -----------------------------------------------------------------

cEcmData::cEcmData(void)
:cECMInfo()
{
  prgId=-1;
}

cEcmData::cEcmData(int PrgId, int Source, int Transponder, cECMInfo *e)
:cECMInfo(e)
{
  prgId=PrgId;
  source=Source;
  transponder=Transponder;
}

bool cEcmData::Parse(const char *buf)
{
  char Name[64];
  Name[0]=0;
  if(sscanf(buf,"%d:%x:%x:%63[^:]:%x:%x:%x\n",&prgId,&source,&transponder,Name,&system,&id,&ecm_pid)!=7)
    return false;
  name=strdup(Name);
  return true;
}

bool cEcmData::Save(FILE *f)
{
  fprintf(f,"%d:%x:%x:%s:%x:%x:%x\n",
            prgId,source,transponder,name,system,id,ecm_pid);
  return ferror(f)==0;
}

// -- cEcmCache ----------------------------------------------------------------

cEcmCache::cEcmCache(void)
:cLoader("ECM")
{}

void cEcmCache::New(int PrgId, int Source, int Transponder, cECMInfo *e)
{
  Lock();
  if(!Exists(PrgId,Source,Transponder,e)) {
    cEcmData *dat=new cEcmData(PrgId,Source,Transponder,e);
    Add(dat);
    Modified();
    d(printf("ecmcache: added prgId=%d source=%x transponder=%x ecm=%x\n",PrgId,Source,Transponder,e->ecm_pid))
    }
  Unlock();
}

bool cEcmCache::Exists(int PrgId, int Source, int Transponder, cECMInfo *e)
{
  cEcmData *dat=Search(PrgId,Source,Transponder);
  return (dat && dat->system==e->system && dat->ecm_pid==e->ecm_pid && dat->id==e->id);
}

cEcmData *cEcmCache::Search(int PrgId, int Source, int Transponder, cEcmData *dat)
{
  Lock();
  if(dat) dat=cList<cEcmData>::Next(dat); else dat=First();
  while(dat) {
    if(dat->prgId==PrgId && dat->source==Source && dat->transponder==Transponder)
      break;
    dat=cList<cEcmData>::Next(dat);
    }
  Unlock();
  return dat;
}

void cEcmCache::Load(void)
{
  Clear();
}

bool cEcmCache::Save(FILE *f)
{
   bool res=true;
   Lock();
   cEcmData *dat=First();
   while(dat) {
     if(!dat->Save(f)) { res=false; break; }
     dat=cList<cEcmData>::Next(dat);
     }
   Modified(!res);
   Unlock();
   return res;
}

bool cEcmCache::ParseLine(char *line, bool fromCache)
{
  bool res=false;
  cEcmData *dat=new cEcmData;
  if(dat && dat->Parse(line)) {
    if(!Exists(dat->prgId,dat->source,dat->transponder,dat)) { Add(dat); dat=0; }
    res=true;
    }
  delete dat;
  return res;
}

// -- cPlainKey ----------------------------------------------------------------

cPlainKey::cPlainKey(void)
{
  au=del=false;
}

bool cPlainKey::Set(int Type, int Id, int Keynr, void *Key)
{
  type=Type; id=Id; keynr=Keynr;
  return SetKey(Key);
}

// -- cHexKey ------------------------------------------------------------------

class cHexKey {
private:
  int keylen;
  void *key;
protected:
  bool HexSetKey(void *Key, int Keylen);
  char *HexPrint(void);
  bool HexCmp(void *Key);
  void HexGet(void *mem);
public:
  cHexKey(void);
  virtual ~cHexKey();
  };

cHexKey::cHexKey(void)
{
  key=0;
}

cHexKey::~cHexKey()
{
  free(key);
}

bool cHexKey::HexSetKey(void *Key, int Keylen)
{
  keylen=Keylen;
  free(key); key=malloc(keylen);
  if(key) memcpy(key,Key,keylen);
  return key!=0;
}

bool cHexKey::HexCmp(void *Key)
{
  return memcmp(key,Key,keylen)==0;
}

void cHexKey::HexGet(void *mem)
{
  memcpy(mem,key,keylen);
}

char *cHexKey::HexPrint(void)
{
  char *str=(char *)malloc(keylen*2+2);
  if(str) HexStr(str,(unsigned char *)key,keylen);
  return str;
}

// -- cBNKey -------------------------------------------------------------------

#if defined(NAGRA_SUPPORT) | defined(CONAX_SUPPORT)

class cBNKey {
private:
  int keylen;
  void *key;
  bool rotate;
protected:
  bool BNSetKey(void *Key, int Keylen);
  char *BNPrint(void);
  bool BNCmp(void *Key);
  void BNGet(void *mem);
public:
  cBNKey(bool Rotate);
  virtual ~cBNKey();
  };

cBNKey::cBNKey(bool Rotate)
{
  rotate=Rotate;
  key=0;
}

cBNKey::~cBNKey()
{
  if(key) BN_free((BIGNUM *)key);
}

bool cBNKey::BNSetKey(void *Key, int Keylen)
{
  keylen=Keylen;
#ifdef NAGRA_SUPPORT
  if(rotate) RotateBytes((unsigned char *)Key,keylen);
#endif
  key=BN_bin2bn((unsigned char *)Key,keylen,0);
  return key!=0;
}

bool cBNKey::BNCmp(void *Key)
{
  return (BN_cmp((BIGNUM *)key,(BIGNUM *)Key)==0);
}

void cBNKey::BNGet(void *mem)
{
  BN_copy((BIGNUM *)mem,(BIGNUM *)key);
}

char *cBNKey::BNPrint(void)
{
  unsigned char mem[keylen*2];
  int len=BN_bn2bin((BIGNUM *)key,mem);
  if(len!=keylen) { // handle short BN
    const int r=keylen-len;
    memmove(mem+r,mem,len);
    memset(mem,0,r);
    }
#ifdef NAGRA_SUPPORT
  if(rotate) RotateBytes(mem,keylen);
#endif
  char *str=(char *)malloc(keylen*2+2);
  if(str) HexStr(str,mem,keylen);
  return str;
}

#endif //NAGRA_SUPPORT || CONAXSUPPORT

// -- cPlain8Key ---------------------------------------------------------------

#define PLAIN8_LEN 8

class cPlain8Key : public cPlainKey, public cHexKey {
protected:
  virtual bool SetKey(void *Key);
  virtual char *Print(void);
  virtual bool Cmp(void *Key);
  virtual void Get(void *mem);
public:
  virtual bool Parse(char *line);
  virtual bool Save(FILE *f);
  };

bool cPlain8Key::SetKey(void *Key)
{
  return HexSetKey(Key,PLAIN8_LEN);
}

bool cPlain8Key::Cmp(void *Key)
{
  return HexCmp(Key);
}

void cPlain8Key::Get(void *mem)
{
  HexGet(mem);
}

char *cPlain8Key::Print(void)
{
  return HexPrint();
}

bool cPlain8Key::Parse(char *line)
{
  unsigned char sid[3], skeynr, skey[PLAIN8_LEN];
#ifdef DEBUG
  char *sline=line;
#endif
  int len;
  if(GetChar(line,&type,1) && (len=GetHex(line,sid,3,false)) &&
     GetHex(line,&skeynr,1) && GetHex(line,skey,PLAIN8_LEN)) {
    type=toupper(type); id=Bin2Int(sid,len); keynr=skeynr;
    SetKey(skey);
    return true;
    }
  d(printf("plain8key: bad key format '%s'\n",sline))
  return false;
}

bool cPlain8Key::Save(FILE *f)
{
  char *str=Print();
  fprintf(f,"%c %.2X %.2X %s\n",type,id,keynr,str);
  free(str);
  return ferror(f)==0;
}

// -- cPlainNKey ---------------------------------------------------------------

#ifdef NAGRA_SUPPORT

#define PLAINN1_LEN 8
#define PLAINN2_LEN 64

class cPlainNKey : public cPlainKey, public cHexKey, public cBNKey {
private:
  bool keyIsBN;
  //
  void SetIsBNKey(void);
protected:
  virtual bool SetKey(void *Key);
  virtual char *Print(void);
  virtual bool Cmp(void *Key);
  virtual void Get(void *mem);
public:
  cPlainNKey(void);
  virtual bool Parse(char *line);
  virtual bool Save(FILE *f);
  };

cPlainNKey::cPlainNKey(void)
:cBNKey(true) // Nagra needs rotated RSA keys
{}

bool cPlainNKey::SetKey(void *Key)
{
  SetIsBNKey();
  if(keyIsBN) return BNSetKey(Key,PLAINN2_LEN);
  else        return HexSetKey(Key,PLAINN1_LEN);
}

void cPlainNKey::SetIsBNKey(void)
{
  keyIsBN=false;
  switch(keynr) {
    case MBC('E','1'):
    case MBC('N','1'):
    case MBC('N','2'): keyIsBN=true; break;
    }
}

bool cPlainNKey::Cmp(void *Key)
{
  if(keyIsBN) return BNCmp(Key);
  else        return HexCmp(Key);
}

void cPlainNKey::Get(void *mem)
{
  if(keyIsBN) BNGet(mem);
  else        HexGet(mem);
}

char *cPlainNKey::Print(void)
{
  if(keyIsBN) return BNPrint();
  else        return HexPrint();
}

bool cPlainNKey::Parse(char *line)
{
  unsigned char sid[2], skeynr;
#ifdef DEBUG
  char *sline=line;
#endif
  int len;
  if(GetChar(line,&type,1) && (len=GetHex(line,sid,2,false))) {
    type=toupper(type); id=Bin2Int(sid,len); 
    line=skipspace(line);
    bool ok;
    switch(toupper(*line)) {
      case 'E':
      case 'N': ok=GetChar(line,&keynr,2); break;

      case 'V': ok=GetChar(line,&keynr,1); break;

      default:  ok=GetHex(line,&skeynr,1); keynr=skeynr; break;
      }
    if(ok) {
      SetIsBNKey();
      const int keylen=keyIsBN ? PLAINN2_LEN : PLAINN1_LEN;
      unsigned char skey[keylen];
      ok=GetHex(line,skey,keylen);
      if(ok) {
        SetKey(skey);
        return true;
        }
      }
    }
  d(printf("nagrakey: bad key format '%s'\n",sline))
  return false;
}

bool cPlainNKey::Save(FILE *f)
{
  char nr[8];
  if(keyIsBN || keynr=='V') {
    int i=(keyIsBN ? 2:1);
    int k=keynr;
    nr[i--]=0; while(i>=0) { nr[i]=k&0xff; k>>=8; i--; }
    }
  else snprintf(nr,sizeof(nr),"%.2X",keynr);

  char *str=Print();
  fprintf(f,"%c %.4X %s %s\n",type,id,nr,str);
  free(str);
  return ferror(f)==0;
}

#endif //NAGRA_SUPPORT

// -- cPlainCKey --------------------------------------------------------------

#ifdef CONAX_SUPPORT

#define PLAINC_LEN 64

class cPlainCKey : public cPlainKey, public cBNKey {
private:
  virtual bool SetKey(void *Key);
  virtual char *Print(void);
  virtual bool Cmp(void *Key);
  virtual void Get(void *mem);
public:
  cPlainCKey();
  virtual bool Parse(char *line);
  virtual bool Save(FILE *f);
  };

cPlainCKey::cPlainCKey()
:cBNKey(false)
{}

bool cPlainCKey::SetKey(void *Key)
{
  return BNSetKey(Key,PLAINC_LEN);
}

bool cPlainCKey::Cmp(void *Key)
{
  return BNCmp(Key);
}

void cPlainCKey::Get(void *mem)
{
  BNGet(mem);
}

char *cPlainCKey::Print(void)
{
  return BNPrint();
}

bool cPlainCKey::Parse(char *line)
{
  unsigned char skeynr, skey[PLAINC_LEN];
#ifdef DEBUG
  char *sline=line;
#endif  
  if(GetChar(line,&type,1) && GetHex(line,&skeynr,1) &&
     GetChar(line,&id,1) && GetHex(line,skey,PLAINC_LEN)) {
    type=toupper(type); keynr=skeynr;
    SetKey(skey);
    return true;
    }
  d(printf("conaxkey: bad key format '%s'\n",sline))
  return false;
}

bool cPlainCKey::Save(FILE *f)
{
  char *str=Print();
  fprintf(f,"%c %.2X %c %s\n",type,keynr,id,str);
  free(str);
  return ferror(f)==0;
}

#endif //CONAX_SUPPORT

// -- cPlainDummyKey -----------------------------------------------------------

class cPlainDummyKey : public cPlainKey {
private:
  virtual bool SetKey(void *Key) { return false; }
  virtual char *Print(void);
  virtual bool Cmp(void *Key) { return false; }
  virtual void Get(void *mem) {}
public:
  virtual bool Parse(char *line) { return false; }
  virtual bool Save(FILE *f) { return false; }
  };

char *cPlainDummyKey::Print(void)
{
  return strdup("dummy");
}

// -- cPlainKeys --------------------------------------------------------------

cPlainKeys::cPlainKeys(void)
:cLoader("KEY")
{}

cPlainKey *cPlainKeys::FindKey(int Type, int Id, int Keynr, cPlainKey *key)
{
  Lock();
  if(key) key=cList<cPlainKey>::Next(key); else key=First();
  while(key) {
    if(!key->IsInvalid() && key->type==Type && key->id==Id && key->keynr==Keynr) break;
    key=cList<cPlainKey>::Next(key);
    }
  Unlock();
  return key;
}

bool cPlainKeys::NewKey(int Type, int Id, int Keynr, void *Key)
{
  cPlainKey *k=0;
  while((k=FindKey(Type,Id,Keynr,k)))
    if(k->Cmp(Key)) return false;

  cPlainKey *nk=NewFromType(Type);
  if(nk) {
    nk->Set(Type,Id,Keynr,Key);
    nk->SetAuto();

    Lock();
    Ins(nk,First());
    Modified();
    Unlock();

    char *str=nk->Print();
    d(printf("key: new key: %c %.2x %02x %s\n",nk->type,nk->id,nk->keynr,str))
    isyslog("key update for ID %c %.2x key number %02x %s",nk->type,nk->id,nk->keynr,str);
    free(str);

/*    k=0;
    while((k=FindKey(Type,Id,Keynr,k))) {
      if(k!=nk && k->IsAuto()) {
        k->Invalid(); modified=true;
        d(printf("key: invalidating key: %c 0x%.2X 0x%02X %s%s\n",k->type,k->id,k->keynr,KeyStr(str,k->key),k->IsAuto()?" (auto)":""))
        }
      }
*/
    return true;
    }
  else esyslog("ERROR: no memory for new softcam key!");
  return false;
}

bool cPlainKeys::Load(const char *cfgdir)
{
  Lock();
  Clear();
  const char *name;
  FILE *f=fopen(name=AddDirectory(cfgdir,KEYFILE),"r");
  if(f) {
    isyslog("loading keys from %s", name);
    char buff[256];
    while(fgets(buff,sizeof(buff),f)) {
      char *line=skipspace(stripspace(buff));
      if(line[0]==0 || line[0]==';') continue; // skip empty & comment lines
      if(!ParseLine(line,false)) { 
        esyslog("ERROR: loading keys failed!");
        d(printf("plainkeys: loading keys failed! Error in line '%s'\n",buff))
        break;
        }
      }
    fclose(f);
    }
  int n=Count();
  d(printf("plainkeys: loaded %d keys from %s\n",n,name))
  if(n) {
#ifdef DEBUG_KEYS
    cPlainKey *dat=First();
    while(dat) {
      if(!dat->IsInvalid()) { printf("plainkeys: "); dat->Save(stdout); }
      dat=cList<cPlainKey>::Next(dat);
      }
    fflush(stdout);
#endif
    }
  Unlock();
  return (n!=0);
}

cPlainKey *cPlainKeys::NewFromType(int type)
{
  cPlainKey *k=0;
  switch(type) {
    case 'I': // Irdeto
    case 'S': // Seca
    case 'V': // Viaccess
      k=new cPlain8Key();
      break;

    case 'C': // Conax
#ifdef CONAX_SUPPORT
      k=new cPlainCKey();
#else
      k=new cPlainDummyKey();
#endif
      break;

    case 'N': // Nagra
#ifdef NAGRA_SUPPORT
      k=new cPlainNKey();
#else
      k=new cPlainDummyKey();
#endif
      break;

    default:
      esyslog("Unknown softcam key type '%c'",type);
      break;
    }
  return k;
}

bool cPlainKeys::ParseLine(char *line, bool fromCache)
{
  char *s=skipspace(line);
  cPlainKey *k=NewFromType(toupper(*s));
  if(k) {
    if(k->Parse((char *)line)) {
      Add(k);
      if(fromCache) k->SetAuto();
      }
    else delete k;
    return true;
    }
  return false;
}

bool cPlainKeys::Save(FILE *f)
{
  bool res=true;
  Lock();
  cPlainKey *dat=First();
  while(dat) {
    if(dat->IsAuto() && !dat->IsInvalid()) {
      if(!dat->Save(f)) { res=false; break; }
      }
    dat=cList<cPlainKey>::Next(dat);
    }
  Modified(!res);
  Unlock();
  return res;
}
