/*
    sio.c
    Serial port input/output functions
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <termios.h>
#include <unistd.h>
#include <errno.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/poll.h>
#include <sys/ioctl.h>
#include <time.h>

#include "sio.h"

#include "logger.h"

#ifdef DEBUG_SIO
#define d(a) a
#else
#define d(a)
#endif

int sio_init(int *i_fd, const char *pc_dev)
{
	struct termios newtio;
	unsigned int modembits;

	d(Log("SIO: Init\n"));

	d(Log("SIO: Open serial port %s\n", pc_dev));

	if ((*i_fd = open(pc_dev, O_RDWR | O_NOCTTY)) < 0)
		return 1;

	d(Log("SIO: Set serial port\n"));

	modembits = TIOCM_DTR;
	if (ioctl(*i_fd, TIOCMBIS, &modembits) < 0)
		return 1;

	modembits = TIOCM_RTS;
	if (ioctl(*i_fd, TIOCMBIC, &modembits) < 0)
		return 1;

	memset(&newtio, 0, sizeof(newtio));
	cfsetispeed(&newtio, B9600);
	cfsetospeed(&newtio, B9600);
//	newtio.c_cflag &= (~PARODD);
//	newtio.c_cflag |= PARENB | CS8 | CSTOPB | CREAD | HUPCL | CLOCAL;
	newtio.c_cflag |= CS8 | CSTOPB | CREAD | HUPCL | CLOCAL;
	newtio.c_iflag &= ~(IGNPAR | PARMRK | INLCR | IGNCR | ICRNL | ISTRIP);
	newtio.c_iflag |= INPCK | BRKINT;
	newtio.c_oflag = 0;
	newtio.c_lflag = 0;
	newtio.c_cc[VMIN] = 1;
	newtio.c_cc[VTIME] = 0;

	if (tcsetattr(*i_fd, TCSANOW, &newtio) < 0)
		return 1;

	if (tcflush(*i_fd, TCIFLUSH) < 0)
		return 1;

	d(Log("SIO: Init done\n"));

	return 0;
}

int  sio_read(int i_fd, int i_timeout, int i_size, uchar * pc_data)
{
	uchar c;
	int i_count, rv;
	struct pollfd pfd;

	d(Log("SIO: Read\n"));

	for (i_count = 0; i_count < i_size; i_count++) {

		pfd.fd = i_fd;
		pfd.events = POLLIN;
		pfd.revents = 0x0000;

		if ((rv = poll(&pfd, 1, i_timeout)) != 1) {
			d(Log("SIO: poll = %d\n", rv));
			return 1;
		}

		if ((pfd.revents & POLLIN) == POLLIN) {
			if (read(i_fd, &c, 1) != 1) {
				d(Log("SIO: Read ERROR\n"));
				return 1;
			}
			pc_data[i_count] = c;
			d(Log("%X ", c));
		} else {
			d(Log("SIO: Read TIMEOUT\n"));
			return 1;
		}
	}

	d(Log("SIO: Read done\n"));

	return 0;
}

int  sio_write(int i_fd, int i_delay, int i_size, const uchar * pc_data)
{
	int i_count;
	struct pollfd pfd;

	d(Log("SIO: Write\n"));

	tcflush(i_fd, TCIFLUSH);

	for (i_count = 0; i_count < i_size; i_count++) {

		if (i_delay > 0)
			usleep(((unsigned long)i_delay) * 1000);

		pfd.fd = i_fd;
		pfd.events = POLLOUT;
		pfd.revents = 0x0000;

		if (poll(&pfd, 1, 1000) != 1) {
			return 1;
		}

		if ((pfd.revents & POLLOUT) == POLLOUT) {
			if (write(i_fd, &pc_data[i_count], 1) != 1) {
				d(Log("SIO: Write ERROR\n"));
				return 1;
			}
			d(Log("%X ", pc_data[i_count]));
		} else {
			d(Log("SIO: Write TIMEOUT\n"));
			/* tcflush (i_fd, TCIFLUSH); */
			return 1;
		}
	}

	d(Log("SIO: Write done\n"));

	return 0;
}

int  sio_close(int i_fd)
{
	d(Log("SIO: Close\n"));

	if (close(i_fd) != 0)
		return 1;

	d(Log("SIO: Close done\n"));

	return 0;
}


int sio_card_inserted(int i_fd)
{
	int status;
	ioctl(i_fd, TIOCMGET, &status);
	
	if (!(status & TIOCM_DSR)) {
		puts("Karte erkannt\n");
	} else {
		puts("keine Karte erkannt\n");
	}

	return (status & TIOCM_DSR) ? 0 : 1;
}



int sio_cardreset(int i_fd)
{
	unsigned int modembits;

	// reset ziehen
	modembits = TIOCM_RTS;
	if (ioctl(i_fd, TIOCMBIS, &modembits) < 0)
		return 1;

	// Buffer flushen
	if (tcflush(i_fd, TCIFLUSH) < 0)
		return 1;

	// 300 ms warten
	usleep(300000);  

	// reset freigeben
	modembits = TIOCM_RTS;
	if (ioctl(i_fd, TIOCMBIC, &modembits) < 0)
		return 1;
}

