/*
    cwemu.c
    Emulate Cryptoworks CAM on Phoenix ISO interface
*/

#include <stdio.h>
#include <unistd.h>
#include <assert.h>
#include <fcntl.h>

#include "sio.h"
#include "iso.h"

//#define ECM_DUMP

#ifdef ECM_DUMP

struct ECMdata {
        int provid;
        int keyid;
        unsigned char chid[5];
        unsigned char ecw[16];
        unsigned char sign[8];
        unsigned char date[11];
};


/* ECM parser code based on vlinder's Yanksee source */
int cw_parse_ecm(unsigned char *data, int len, struct ECMdata *ed)
{
  unsigned char *source=data+5; // Skip 81 FF 00 00 (Len)
  int length=len-5;
  unsigned char *p=source;

  while(p-source-2<length)
  {
    switch(*p)
    {
      case 0x83: // ID+Key PPPPPkkk
        ed->provid=p[2]&0xF8;
        ed->keyid=p[2]&0x07;
        break;
      case 0x8C: // Status +CHID
        sprintf(ed->chid,"%04X",p[3]*255+p[4]);
        break;
      case 0x8E: // Date
        sprintf(ed->date,"%02d/%02d/%04d",p[3]&0x1F, (p[3]>>5)+(p[2]&1)*8,
					  1990+(p[2]>>1));
        break;
      case 0x8F: // ??
        break;
      case 0x91: // ??
        break;
      case 0xC5: // ??
        break;
      case 0xDB: // encrypted CW
        memcpy(ed->ecw,p+2,16);
        break;
      case 0xDF: // signature
        memcpy(ed->sign,p+2,8);
        break; 
    }
    p+=2+p[1]; // p[1]=sub block len
  }

  return p-data;
}

int cw_parse_atr(unsigned char *data) {

	unsigned char *pc_atr = data;

	if(!memcmp(pc_atr, "\x3B\x78\x12\x0\x0\x54\xC4", 7)) {
		// OK
		printf("CW card, v.%d (pindown=%d)\n", pc_atr[7], pc_atr[8]);
		if(pc_atr[9] != 0x8F || pc_atr[10] != 0xF1)
			printf("warning: bad ATR postfix: %2X %2X\n", pc_atr[9], pc_atr[10]);
		return 0;

	}
	return -1;
}

int cw_get_cardno(unsigned char *serno) {

	unsigned char pc_atr[MAX_LEN];
	
	if(!i_iso_write("\xA4\xA4\x0\x0\x2", "\x2F\x1", "\x9F\x11"))
	  if(!i_iso_write("\xA4\xA2\x0\x0\x1", "\x80", "\x9F\x7"))
	    if(!i_iso_read("\xA4\xB2\x0\x0\x7", pc_atr, "\x90\x0")) {
	      memcpy(serno, pc_atr+2, 5);
	      return 0;
	    }
	memset(serno, 0, 5);
	return -1;
}

void v_dump(int i_len, unsigned char *pc_data); // from sio.h

int cw_get_perms() {

	unsigned char pc_atr[MAX_LEN];
	
	if(!i_iso_write("\xA4\xA4\x0\x0\x2", "\x1F\x50", "\x9F\x11"))
	  if(!i_iso_write("\xA4\xA4\x0\x0\x2", "\x0F\x20", "\x9F\x11"))
	    if(!i_iso_write("\xA4\xA2\x1\x0\x5", "\x8C\x0\x0\x0\x0", "\x9F\x42"))
	    if(!i_iso_read("\xA4\xB2\x0\x0\x42", pc_atr, "\x90\x0")) {
	      printf("Packet: '%s'\n", pc_atr+12);
	      v_dump(12, pc_atr+12);
	}
	return 0;
}

void cw_ecm_dump(unsigned char *data) {

	int len = data[4]+5-1, i;
	struct ECMdata ed;

	v_dump(20, data);
	if(cw_parse_ecm(data, len, &ed)) {
	   printf(" CW: %02X %02X %-4s %s\n",
		   ed.provid, ed.keyid, ed.chid, ed.date);
           printf("     ");
           for (i=0; i<sizeof(ed.ecw); i++)
             printf(" %02x",ed.ecw[i]);
           printf(" /");
           for (i=0; i<sizeof(ed.sign); i++)
             printf(" %02x",ed.sign[i]);
           printf("\n");
	 }
}

#endif /*ECM_DUMP*/

int cryptoworks_decode(unsigned char *cws, unsigned char *data, char *portname)
{
  static init = 1;
  static unsigned char pc_insW[] = { 0xA4,0x4C,0x00,0x00,0x00 };
  static unsigned char pc_insR[] = { 0xA4,0xC0,0x00,0x00,0x00 };
  static unsigned char pc_svW[]  = { 0x9F,0x1C };
  static unsigned char pc_svR[]  = { 0x90,0x00 };
  unsigned char pc_get[MAX_LEN];

#ifdef ECM_DUMP
  cw_ecm_dump(data+1+3);
#endif

  if(init) {
    if(i_iso_open(portname, pc_get)) return 1;
    init = 0;
#ifdef ECM_DUMP
    if(!cw_parse_atr(pc_get)) {
      unsigned char cardno[5];
      cw_get_cardno(cardno);
      printf("Card SerNo:");
      for(i=0; i < 5; i++) printf(" %02X", cardno[i]);
      printf("\n");
      cw_get_perms();
      }
#endif
    }

  pc_insW[4]=data[2]-2;
  if(i_iso_write(pc_insW,data+5,pc_svW)) return 1;
  pc_insR[4]=pc_svW[1];
  if(i_iso_read(pc_insR,pc_get,pc_svR)) return 1;
  memcpy(cws,pc_get+2,16);
  return 0;
}
