/*
 * Softcam plugin to VDR (C++)
 *
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 */

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>

#include "common.h"
#include "logger-irdeto.h"
#include "logger.h"
#include "softcam.h"
#include "data.h"
#include "misc.h"

// -- cIrdCardInfo ----------------------------------------------------------------

class cIrdCardInfo : public cListObject {
private:
  bool updated;
public:
  int Provider;
  unsigned char ID[3], PMK[8], HS[3], HMK[10];
  bool haveHS, haveHMK;
  //
  cIrdCardInfo(void);
  bool Parse(char *line);
  bool Save(FILE *f);
  bool Update(unsigned char *pmk, unsigned char *id);
  bool IsUpdated(void) { return updated; }
  void Updated(void) { updated=true; }
  bool Cmp(cIrdCardInfo *ci);
  };

cIrdCardInfo::cIrdCardInfo(void)
{
  updated=false;
}

bool cIrdCardInfo::Parse(char *line)
{
  unsigned char p[1];
  while(*line && isspace(*line)) line++;
  if(line[0]!=';') {
    haveHS=haveHMK=false;
    if(GetHex(line,HS,sizeof(HS))) haveHS=true;
    if(GetHex(line,HMK,sizeof(HMK))) haveHMK=true;
    if(GetHex(line,p,sizeof(p)) && GetHex(line,ID,sizeof(ID)) && GetHex(line,PMK,sizeof(PMK))) {
      Provider=(int)p[0];
      return true;
      }
    }
  return false;
}

bool cIrdCardInfo::Save(FILE *f)
{
  char s1[20], s2[20], s3[20], s4[20];
  fprintf(f,"%s %s %02x %s %s\n",
     HexStr(s1,HS,sizeof(HS)),HexStr(s2,HMK,sizeof(HMK)),Provider,
     HexStr(s3,ID,sizeof(ID)),HexStr(s4,PMK,sizeof(PMK)));
  return ferror(f)==0;
}

bool cIrdCardInfo::Cmp(cIrdCardInfo *ci)
{
  return (!memcmp( HS,ci->HS ,sizeof( HS)) &&
          !memcmp(HMK,ci->HMK,sizeof(HMK)) );
}

bool cIrdCardInfo::Update(unsigned char *pmk, unsigned char *id)
{
  bool res=false;
  char str[20], str2[12];
  if(memcmp(PMK,pmk,sizeof(PMK))) {
    isyslog("new PMK for I card %s: %s",HexStr(str2,HS,sizeof(HS)),KeyStr(str,pmk));
    memcpy(PMK,pmk,sizeof(PMK));
    Updated(); res=true;
    }
  if(id && memcmp(ID,id,sizeof(ID))) {
    isyslog("new PrvID for I card %s: %s",HexStr(str2,HS,sizeof(HS)),HexStr(str,id,sizeof(ID)));
    memcpy(ID,id,sizeof(ID));
    Updated(); res=true;
    }
  return res;
}

// -- cIrdCardInfos ---------------------------------------------------------------

class cIrdCardInfos : public cCardInfos<cIrdCardInfo> {
public:
  cIrdCardInfos(void);
  bool Update(cIrdCardInfo *ci, unsigned char *pmk, unsigned char *id);
  };

cIrdCardInfos *Icards=0;

cIrdCardInfos::cIrdCardInfos(void)
:cCardInfos<cIrdCardInfo>("IrdBeta")
{}

bool cIrdCardInfos::Update(cIrdCardInfo *ci, unsigned char *pmk, unsigned char *id)
{
  bool res=ci->Update(pmk,id);
  if(res) Modified();
  return res;
}

// -- cLogIrdeto ------------------------------------------------------------------

cLogIrdeto::cLogIrdeto(int CardNum)
:cLogSys(CardNum)
{
  lastKey=0;
}

bool cLogIrdeto::HaveCards(void)
{
  return Icards->HaveCards();
}

bool cLogIrdeto::LoadCards(const char *cfgdir)
{
  if(!Icards) Icards=new cIrdCardInfos;
  if(!Icards) {
    esyslog("ERROR: no memory for Ird cards!");
    return false;
    }
  Icards->Load(cfgdir,"Irdeto","Ird-Beta.KID");
  return true;
}

int cLogIrdeto::ParseCAT(unsigned char *buffer, int n, int sys, int id)
{
  for(int index=8 ; index<n ;) {
    if(buffer[index]==0x09) {
      int s=WORD(buffer,index+2,0xFFFF);
      if(sys==s) return WORD(buffer,index+4,0x1FFF);
      }
    index+=buffer[index+1]+2;
    }
  return -1;
}

int cLogIrdeto::Section(void)
{
  return 0x82;
}

void cLogIrdeto::Log(unsigned char *buffer, int n)
{
  int i, numKeys=0, date=0;
  unsigned char adr[4]={0,0,0,0}, id[4], *pk[4], prov, *mk=0, prvId[3]={0,0,0};

  unsigned char savebuf[4096];
  if(n>(int)sizeof(savebuf)) {
    dl(printf("logger %d: paket size %d to big for savebuffer in IrdetoLog\n",cardNum,n))
    return;
    }

  unsigned char adrT=buffer[3];
  int index=adrT&7;
  for(i=0 ; i<index ; i++) adr[i]=buffer[4+i];
  index+=4+4; // 3 header + adr type + 4 {0x01 0x00 0x00 len}
  
#ifdef EMM_DUMP
  switch(adrT) {
    case 0x02:
    case 0x03:
    case 0x0a:
    case 0x0b:
    case 0xc3: break;
    default:   Dump(EMM_DUMP"_unkown_cmd",buffer,n);
               break;
    }
#endif

#ifdef EMM_DUMP
  if(adrT==0xc3) Dump(EMM_DUMP"_c3",buffer,n);
#endif

  int sindex=index;           // save index for sig. check
  int slen=buffer[index-1]-5; // save packet length, 5 signature bytes
  int maxIndex=index+slen;
  if(maxIndex>n-5) {
    dl(printf("logger %d: bad packet length (%d > %d)\n",cardNum,maxIndex,n-5))
    maxIndex=n-5;
    }
  bool badnano=false;
  while(!badnano && index<maxIndex) {
    unsigned char nlen=buffer[index+1] & 0x3F;
    //unsigned char prio=buffer[index] & 0x40;
    unsigned char nano=buffer[index] & ~0x40;
    switch(nano) {
      case 0x10: // key update
        {
        int k=(buffer[index+1]>>6)+1; // key counter
        if(nlen!=k*9) { badnano=true; break; }
        for(i=0 ; i<k ; i++) {
          id[i]= buffer[index+2+0+i*9];
          pk[i]=&buffer[index+2+1+i*9];
          numKeys++;
          }
        }
	break;
      case 0x00: // date
        if(nlen<2) { badnano=true; break; }
	date=WORD(buffer,index+2,0xFFFF);
	break;
      case 0x28: // pmk & provid update
        if(nlen!=13) { badnano=true; break; }
	prov= buffer[index+2+0];
	mk=  &buffer[index+2+2];
	prvId[0]=buffer[index+2+10];
        prvId[1]=buffer[index+2+11];
        prvId[2]=buffer[index+2+12];
        break;
      case 0x29: // pmk update
        if(nlen!=10) { badnano=true; break; }
	prov= buffer[index+2+0];
	mk=  &buffer[index+2+2];
	break;

      case 0x11: // channel id
        if(nlen!=6) { badnano=true; break; }
        //chId[0]=buffer[index+2+0];
        //chId[1]=buffer[index+2+1];
        break;
      case 0x91: // erase channel id
        if(nlen!=6) { badnano=true; break; }
        //eraseChId[0]=buffer[index+2+0];
        //eraseChId[1]=buffer[index+2+1];
        break;
      case 0x8B: // CB20-matrix
        if(nlen!=0x20) { badnano=true; break; }
        //cb20ptr=&buffer[index+2];
        break;
      case 0x22: // set country code
        if(nlen!=3) { badnano=true; break; }
        //
        break;
      case 0x95: // unknown
        if(nlen!=2) { badnano=true; break; }
        //
        break;
      case 0x1E: // unknown
        if(nlen!=15) { badnano=true; break; }
        //
        break;
      case 0x1F: // unknown
        if(nlen!=3) { badnano=true; break; }
        //
        break;
      case 0x16: // unknown
        if(nlen!=2) { badnano=true; break; }
        //
        break;
      case 0x12: // unknown
        if(nlen!=6) { badnano=true; break; }
        //
        break;

      default:
        dl(printf("logger %d: unhandled nano 0x%02x\n",cardNum,nano))
#ifdef EMM_DUMP
        Dump(EMM_DUMP"_unhandled_nano",buffer,n);
#endif
        break;
      }
    index+=nlen+2;
    }
  if(badnano || index!=maxIndex) {
    dl(printf("logger %d: bad nano/bad paket\n",cardNum))
#ifdef EMM_DUMP
    Dump(EMM_DUMP"_bad",buffer,n);
#endif
    return;
    }

  // lastKey: save cpu time if we get bursts of the same key
  if((numKeys>0 && (id[0]!=lastKey || numKeys>1)) || mk) {
    memcpy(savebuf,buffer,n); // save the buffer
    cIrdCardInfo *ci=Icards->First();
    while(ci) {
      if((numKeys>0                       && ((adr[0]==ci->ID[0] && adr[1]==ci->ID[1]) || (!ci->ID[0] && !ci->ID[1] && !ci->ID[2]))) ||
         (mk && ci->haveHMK && ci->haveHS && (adr[0]==ci->HS[0] && adr[1]==ci->HS[1] && adr[2]==ci->HS[2]))) {
        dl(printf("%02x ID %02x %02x %02x",adrT,adr[0],adr[1],adr[2]))

        for(i=0 ; i<numKeys ; i++) {
          lastKey=id[i];
          sessionKeyCrypt(pk[i],ci->PMK,date);
          }
        unsigned char chkkey[max(sizeof(ci->PMK),sizeof(ci->HMK))];
        int keylen;
        if(mk) {
          dl(printf("PMK update -> check sign\n")) 
          memcpy(chkkey,ci->HMK,sizeof(ci->HMK)); // key is modified in decryptIrd()
          decryptIrd(mk,chkkey,128,16);
          keylen=sizeof(ci->HMK);
          memcpy(chkkey,ci->HMK,sizeof(ci->HMK)); // signature check with HMK
          }
       else {
          keylen=sizeof(ci->PMK);
          memcpy(chkkey,ci->PMK,sizeof(ci->PMK)); // signature check with PMK
          }

#ifdef EMM_DUMP
        Dump(EMM_DUMP,buffer,n);
#endif
        buffer[sindex-6]=adrT;
        buffer[sindex-5]=adr[0]; buffer[sindex-4]=adr[1]; buffer[sindex-3]=adr[2]; buffer[sindex-2]=adr[3];
        if(signatureCheck(&buffer[sindex-6],slen+6,chkkey,date,&buffer[sindex+slen],keylen))
	  {
          char str[20];
#ifdef DEBUG_LOG
          printf(" - OK");
          for(i=0 ; i<numKeys ; i++) printf(" - PK %02x %s",id[i],KeyStr(str,pk[i]));
          if(mk) printf(" - PMK %s",KeyStr(str,mk));
          printf("\n");
#endif

          for(i=0 ; i<numKeys ; i++) {
            cSoftCAM::FoundKey();
            if(keys->NewKey('I',ci->Provider,id[i],pk[i])) {
              cSoftCAM::NewKey();
#ifdef DEBUG_LOG_FILE
              FILE *f=LogStart();
              if(f) {
                fprintf(f,"(%02x) %02x %02x%02x%02x - PK %02x %s\n",adrT,ci->Provider,adr[0],adr[1],adr[2],id[i],KeyStr(str,pk[i]));
                LogEnd(f);
                }
#endif
              }
            }
          if(mk) {
            cSoftCAM::FoundKey();
            if(Icards->Update(ci,mk,prvId[0]?prvId:0)) {
              cSoftCAM::NewKey();
#ifdef DEBUG_LOG_FILE
              FILE *f=LogStart();
              if(f) {
                fprintf(f,"(%02x) %02x %02x%02x%02x - PMK %s with ID %02x%02x%02x\n",adrT,ci->Provider,adr[0],adr[1],adr[2],KeyStr(str,mk),prvId[0],prvId[1],prvId[2]);
                LogEnd(f);
                }
#endif
              }
            }
          cSoftCAM::Save();
          break;
          }
        else {
          dl(printf(" - FAIL\n"))
#ifdef DEBUG_LOG_FILE
          FILE *f=LogStart();
          if(f) {
            fprintf(f,"(%02x) %02x %02x%02x%02x - FAIL %s\n",adrT,ci->Provider,adr[0],adr[1],adr[2],keylen==8?"PK":"PMK");
            LogEnd(f);
            }
#endif
          }

        memcpy(buffer,savebuf,n); // restore the buffer
        }
      ci=Icards->cList<cIrdCardInfo>::Next(ci);
      }
    }
}
