/*
 * Softcam plugin to VDR (C++)
 *
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 */

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>

#include "common.h"
#include "logger.h"
#include "logger-irdeto.h"
#include "logger-seca.h"
#include "logger-nagra.h"
#include "logger-betacam.h"
#include "softcam.h"
#include "data.h"
#include "setup.h"
#include "misc.h"

// -- cLogger ------------------------------------------------------------------

cLogger::cLogger(int CardNum, int EmmDev)
:cRunMode("logger",CardNum)
{
  emmDev=EmmDev;
  cardNum=CardNum;
  caSys=newCaSys=Id=newId=0; currentLogger=0;
  Start();
  WaitRunMode(rmStartup,true);
}

cLogger::~cLogger()
{
  if(runMode!=rmKilled) {
    SetCaSys(0,0);
    SetRunMode(rmStop);
    WaitRunMode(rmKilled,false);
    }
  delete currentLogger;
}

void cLogger::SetCaSys(int CaSys, int ID)
{
  WaitRunMode(rmChange,true);
  if(CaSys!=caSys || ID!=Id) {
    d(printf("logger %d: setting new CA_SYS 0x%04x ID 0x%04x\n",cardNum,CaSys,ID))
    newCaSys=CaSys; newId=ID;
    SetRunMode(rmChange);
    if(CaSys) {
      WaitRunMode(rmChange,true,5000);
      if(runMode==rmChange && runMode!=rmKilled) d(printf("logger %d: timeout setting CA_SYS\n",cardNum))
      }
    }
}

bool cLogger::LoadCards(const char *cfgdir)
{
  if(cLogIrdeto::LoadCards(cfgdir) &&
     cLogSeca::LoadCards(cfgdir))
    return true;
  return false;
}

int cLogger::ParseCAT(void)
{
  int pid=-1;
  SetFilter(emmDev,1,true,0x01,0xff,0);
  for(int count=PAT_TRY_COUNT ; count>0 ; count--) {
    unsigned char buffer[4096];
    int n=ReadPaket(emmDev,buffer,sizeof(buffer));
    if(n==-1) break;
    if(n>0) {
#ifdef CAT_DUMP
      Dump(CAT_DUMP,buffer,n);
#endif
      if(buffer[0]==0x01) {
        n-=4;  // garbage at end of packet
        pid=currentLogger->ParseCAT(buffer,n,caSys,Id);
        if(pid>=0) break;
        }
      }
    }
  StopFilter(emmDev);
  if(pid<0) d(printf("logger %d: no EMM pid for CA_SYS 0x%04x ID 0x%04x in CAT\n",cardNum,caSys,Id))
  return pid;
}

void cLogger::Action(void)
{
  dsyslog("logger thread started on card %d (pid=%d)", cardNum, getpid());
  nice(3); // don't disturb the softcam thread
  SetRunMode(rmRun);

  while(runMode!=rmStop) {
    delete currentLogger; currentLogger=0;
    WaitRunMode(rmChange,false);

    if(runMode==rmChange) { caSys=newCaSys; Id=newId; SetRunMode(rmRun); }
    switch(caSys & MASK_CA_SYSTEM) {
#ifdef BETACAM_SUPPORT
      case IRDETO_CA_SYSTEM:
        currentLogger=new cLogIrdeto(cardNum);
				break;
      case BETA_CA_SYSTEM:
        currentLogger=new cLogBetaCam(cardNum);
        break;
#else  // BETACAM_SUPPORT
      case IRDETO_CA_SYSTEM:
      case BETA_CA_SYSTEM:
        currentLogger=new cLogIrdeto(cardNum);
        break;
#endif // BETACAM_SUPPORT
      case SECA_CA_SYSTEM:
        currentLogger=new cLogSeca(cardNum);
        break;
      //case VIACCESS_CA_SYSTEM:
#ifdef NAGRA_SUPPORT
      case NAGRA_BEV_CA_SYSTEM:
      case NAGRA_CA_SYSTEM:
        currentLogger=new cLogNagra(cardNum);
        break;
#endif 
      }
    if(!caSys || !ScSetup.LoggerActive || !currentLogger || !currentLogger->HaveCards()) { caSys=0; continue; }

    if(ScSetup.UseFixedEMMPid) emmPid=ScSetup.FixedEMMPid;
    else emmPid=ParseCAT();
    if(emmPid<0) { caSys=0; continue; }

    int section=currentLogger->Section();
    d(printf("logger %d: start logging for CA_SYS 0x%04x ID 0x%04x on EMM pid 0x%04x (section 0x%02x)\n",cardNum,caSys,Id,emmPid,section))
    SetFilter(emmDev,emmPid,false,section,0xff,32768);

    int stopTime=0;
    if(ScSetup.LoggerTimeout) stopTime=time(0)+ScSetup.LoggerTimeout;

    while(runMode==rmRun && ScSetup.LoggerActive && (!stopTime || time(0)<stopTime)) {
      unsigned char buffer[4096];
      int n=ReadPaket(emmDev,buffer,sizeof(buffer),SYNC_TIMEOUT);
      if(n>0 && buffer[0]==section) currentLogger->Log(buffer,n);
      }
    if(runMode!=rmChange) d(printf("logger %d: stop logging for CA_SYS 0x%04x ID 0x%04x on EMM pid 0x%04x\n",cardNum,caSys,Id,emmPid))
    StopFilter(emmDev); caSys=0;
    }

  SetRunMode(rmKilled);
  dsyslog("logger thread ended on card %d (pid=%d)", cardNum, getpid());
}
