/*
 * Softcam plugin to VDR (C++)
 *
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 */

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/poll.h>
#include <sys/time.h>
#include <unistd.h>

#include <linux/dvb/dmx.h>

#include <vdr/tools.h>
#include <vdr/thread.h>

#include "common.h"
#include "misc.h"

// -----------------------------------------------------------------------------

void SetFilter(int fd, int pid, bool crc, int tnr, int mask, int buffSize)
{
  if(buffSize<8192) buffSize=8192;
  CHECK(ioctl(fd,DMX_SET_BUFFER_SIZE,buffSize));

  struct dmx_sct_filter_params FilterParams;
  memset(&FilterParams,0,sizeof(FilterParams));
  FilterParams.filter.filter[0]=tnr;
  FilterParams.filter.mask[0]=mask;
  FilterParams.flags=DMX_IMMEDIATE_START; if(crc) FilterParams.flags|=DMX_CHECK_CRC;
  FilterParams.pid=pid;
  CHECK(ioctl(fd,DMX_SET_FILTER,&FilterParams));
}

void StopFilter(int fd)
{
  CHECK(ioctl(fd,DMX_STOP));
}

int TimedRead(int fd, unsigned char *buffer, int length, int timeout)
{
  int r=read(fd,buffer,length);
  if(r<0 && errno==EAGAIN) {
    struct pollfd u;
    u.fd=fd; u.events=POLLIN;
    r=poll(&u,1,timeout);
    if(r<0) {
      d(printf("TimedRead: poll: %s\n",strerror(errno)))
      return -1;
      }
    else if(r==0) return -2; // timeout
    r=read(fd,buffer,length);
    if(r<0)
      {
      if(errno!=EOVERFLOW) d(printf("TimedRead: read: Buffer overflow\n"))
      else d(printf("TimedRead: read: %s\n",strerror(errno)))
      if(errno==EAGAIN) r=0;
      }
    }
  return r;
}

int ReadPaket(int fd, unsigned char *buffer, int length, int timeout)
{
  int n=0, size=length;
#ifdef DEBUG
  int i=0;
#endif
  do {
    int nn=TimedRead(fd,buffer+n,size-n,timeout);
#ifdef DEBUG
    i++;
#endif
    if(nn<0) return nn;
    n+=nn;
    if(size==length && n>=3) {
      size=WORD(buffer,1,0x0FFF)+3;
      if(size>length) {
        d(printf("ReadPaket: buffer size %d too small for paket %d\n",length,size))
        return -1;
        }
      }
    } while(n<3 || n<size);
#ifdef DEBUG
  if(i>1) printf("ReadPaket: n=%d size=%d : %d reads\n",n,size,i);
  if(n!=size) printf("ReadPaket: n=%d size=%d : size mismatch\n",n,size);
  if(size>188) printf("ReadPaket: n=%d size=%d : big paket\n",n,size);
#endif
  return n;
}

void Dump(const char *name, unsigned char *buffer, int n)
{
#ifdef DEBUG
  FILE *f=fopen(name,"a");
  if(f) {
    fprintf(f,"read: n=%d/0x%04x n-3=%04x\n",n,n,n-3);
    for(int i=0 ; i<n ; i+=8) {
      fprintf(f,"%04x: ",i);
      for(int l=0 ; l<8 && i+l<n ; l++) fprintf(f,"%02x ",buffer[i+l]);
      fprintf(f,"\n");
      }
    fclose(f);
    }
#endif
}

void HexDump(unsigned char *buffer, int n)
{
#ifdef DEBUG
  printf("read: n=%d/0x%04x n-3=%04x\n",n,n,n-3);
  for(int i=0 ; i<n ; i+=8) {
    printf("%04x: ",i);
    for(int l=0 ; l<8 && i+l<n ; l++) printf("%02x ",buffer[i+l]);
    printf("\n");
    }
#endif
}

#ifdef DEBUG
static cMutex filelog;
#endif

FILE *LogStart(const char *name)
{
  FILE *f=0;
#ifdef DEBUG
#ifdef DEBUG_LOG_FILE
  if(!name) name=DEBUG_LOG_FILE;
#endif

  if(name) {
    filelog.Lock();
    f=fopen(name,"a");
    if(f) {
      time_t t=time(0);
      char *tt=strdup(ctime(&t));
      int l=strlen(tt);
      if(l) tt[l-1]=0;
      fprintf(f,"%s: ",tt);
      free(tt);
      }
    else filelog.Unlock();
    }
#endif
  return(f);
}

void LogEnd(FILE *f)
{
#ifdef DEBUG
  fflush(f); fclose(f);
  filelog.Unlock();
#endif
}

const char *HexStr(char *str, const unsigned char *mem, int len)
{
  for(int i=0 ; i<len ; i++) sprintf(&str[i*2],"%02X",mem[i]);
  return str;
}

static int HexDigit(int asc)
{
  if(asc>='A') return toupper(asc)-'A'+10;
  return asc-'0';
}

int GetChar(char * &line, int *store, int count)
{
  line=skipspace(line);
  char *sline=line;
  *store=0;
  while(count) {
    if(line[0]==0 && !isalnum(line[0])) break;
    *store<<=8; *store+=line[0];
    line++; count--;
    }
  if(!count && (!*line || isspace(*line) || *line==';')) return sline-line;
  *store=0;
  line=sline;
  return 0;
}

static int GetHexBase(char * &line, unsigned char *store, int count, bool fixedLen, bool asc)
{
  line=skipspace(line);
  char *sline=line;
  unsigned char *sstore=store;
  int scount=count;
  while(count) {
    if(line[0]==0 || line[1]==0 || !isxdigit(line[0]) || !isxdigit(line[1])) break;
    if(asc) {
       *store++=line[0]; *store++=line[1];
       }
    else {
      int d1=HexDigit(line[0]);
      int d0=HexDigit(line[1]);
      *store++=(d1<<4) + d0;
      }
    line+=2; count--;
    }
  if(asc) *store=0; // make sure strings are NULL terminated
  if((!count || (!fixedLen && count!=scount)) &&
     (!*line || isspace(*line) || *line==';') ) return scount-count;
  memset(sstore,0,scount);
  line=sline;
  return 0;
}

int GetHex(char * &line, unsigned char *store, int count, bool fixedLen)
{
  return GetHexBase(line,store,count,fixedLen,false);
}

int GetHexAsc(char * &line, unsigned char *store, int count)
{
  return GetHexBase(line,store,count/2,true,true)*2;
}

int Bin2Int(unsigned char *mem, int len)
{
  int k=0;
  for(int i=0 ; i<len ; i++) { k<<=8; k+=mem[i]; }
  return k;
}

// -- cRunMode -----------------------------------------------------------------

cRunMode::cRunMode(const char *Module, const int Card)
{
  runMode=rmStartup;
  dc(snprintf(module,sizeof(module),"%s %d",Module,Card))
}

void cRunMode::SetRunMode(eRunMode mode)
{
  runModeMutex.Lock();
  runMode=mode;
  dc(printf("%s: setting mode=%d (pid=%d)\n",module,mode,getpid()))
  runModeCond.Broadcast();
  runModeMutex.Unlock();
}

void cRunMode::WaitRunMode(eRunMode mode, bool inv, unsigned long timeout)
{
  runModeMutex.Lock();
  while(runMode!=rmStop && runMode!=rmKilled && ((!inv && mode!=runMode) || (inv && mode==runMode))) {
    dc(printf("%s: entering wait for mode%s%d with mode=%d timeout=%ld (pid=%d)\n",module,inv?"!=":"==",mode,runMode,timeout,getpid()))
    if(timeout) {
      if(!runModeCond.TimedWait(runModeMutex,timeout)) {
        // timeout
        dc(printf("%s: timeout while waiting (mode=%d) (pid=%d)\n",module,runMode,getpid()))
        break;
        }
      }
    else 
      runModeCond.Wait(runModeMutex);

    dc(printf("%s: returning from wait with mode=%d (pid=%d)\n",module,runMode,getpid()))
    }
  runModeMutex.Unlock();
}
