/*
 * Softcam plugin to VDR (C++)
 *
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 */

#include <malloc.h>
#include <stdlib.h>

#include <vdr/plugin.h>
#include <vdr/menuitems.h>

#include "common.h"
#include "setup.h"
#include "softcam.h"
#include "status.h"
#include "i18n.h"

#define MIN_VERS   1 // required VDR version
#define MIN_MAJOR  1
#define MIN_MINOR 31

static const char *VERSION        = "0.2.2";
static const char *DESCRIPTION    = "A software emulated CAM";

// --- cMenuSetupSc ------------------------------------------------------------

class cMenuSetupSc : public cMenuSetupPage {
private:
  cScSetup data;
  char *cfgdir;
  //
  cOsdItem *Item(const char *name, const char *text);
  cOsdItem *Item(const char *name, const int num);
protected:
  virtual void Store(void);
public:
  cMenuSetupSc(const char *CfgDir);
  virtual ~cMenuSetupSc();
  virtual eOSState ProcessKey(eKeys Key);
  };

cMenuSetupSc::cMenuSetupSc(const char *CfgDir)
{
  static const char *logg[3];
  logg[0]=tr("off");
  logg[1]=tr("always on");
  logg[2]=tr("non-recording on");

  cfgdir=strdup(CfgDir);
  data=ScSetup;
  SetSection(tr("SoftCAM"));
  Add(new cMenuEditStraItem(  tr("Key logger")        ,&data.LoggerActive,3,logg));
  Add(new cMenuEditIntItem(   tr("Key logger timeout"),&data.LoggerTimeout, 0, 9999));
  Add(new cMenuEditBoolItem(  tr("Use fixed EMM PID") ,&data.UseFixedEMMPid));
  Add(new cMenuEditIntItem(   tr("Fixed EMM PID"),     &data.FixedEMMPid, 3, 0x1ffe));
  Add(new cOsdItem(           tr("Reload files"),      osUser1));
  int fk, nk;
  cSoftCAM::KeyStats(fk,nk);
  Add(Item(tr("[Found keys]"),fk));
  Add(Item(tr("[New keys]"),  nk));
  for(int d=0; d<MAXDVBAPI; d++) {
    const char *ks=cSoftCAM::CurrKeyStr(d);
    if(ks) {
      char buffer[32];
      snprintf(buffer,sizeof(buffer),"[%s %d]",tr("Current key card"),d+1);
      Add(Item(buffer,ks));
      }
    }
}

cMenuSetupSc::~cMenuSetupSc()
{
  free(cfgdir);
}

cOsdItem *cMenuSetupSc::Item(const char *name, const int num)
{
  char buf[32];
  snprintf(buf,sizeof(buf),"%d",num);
  return Item(name,buf);
}

cOsdItem *cMenuSetupSc::Item(const char *name, const char *text)
{
  char *buf=0;
  asprintf(&buf,"%s:\t%s",name,text);
  cOsdItem *item = new cOsdItem(buf,osBack);
  item->SetColor(clrWhite, clrBackground);
  free(buf);
  return item;
}

void cMenuSetupSc::Store(void)
{
  ScSetup=data;
  // remove old setup entries
  SetupStore("XorHack");
  SetupStore("FirmwareHack");
  // save current setup
  SetupStore("LoggerActive",   ScSetup.LoggerActive);
  SetupStore("LoggerTimeout",  ScSetup.LoggerTimeout);
  SetupStore("UseFixedEMMPid", ScSetup.UseFixedEMMPid);
  SetupStore("FixedEMMPid",    ScSetup.FixedEMMPid);
}

eOSState cMenuSetupSc::ProcessKey(eKeys Key)
{
  eOSState state = cOsdMenu::ProcessKey(Key);

  switch(state) {
    case osUser1:
      state=osContinue;
      if(!cSoftCAM::Active()) {
        if(Interface->Confirm(tr("Really reload files?"))) {
          Store();
          cSoftCAM::Startup(cfgdir);
          state=osEnd;
          }
        }
      else Interface->Error(tr("Active! Can't reload files now"));
      break;
    case osUnknown:
      switch (Key) {
        case kOk: Store();
                  state=osBack;
                  break;
        default: break;
        }
      break;
    default:
      break;
    }
  return state;
}

// --- cPluginSc ---------------------------------------------------------------

class cPluginSc : public cPlugin {
private:
  cScStatus *status;
  //
  bool CheckVersion(void);
public:
  cPluginSc(void);
  virtual ~cPluginSc();
  virtual const char *Version(void) { return VERSION; }
  virtual const char *Description(void) { return DESCRIPTION; }
  virtual bool Initialize(void);
  virtual cMenuSetupPage *SetupMenu(void);
  virtual bool SetupParse(const char *Name, const char *Value);
  };

cPluginSc::cPluginSc(void)
{
  // Initialize any member variables here.
  // DON'T DO ANYTHING ELSE THAT MAY HAVE SIDE EFFECTS, REQUIRE GLOBAL
  // VDR OBJECTS TO EXIST OR PRODUCE ANY OUTPUT!
  status=0;
}

cPluginSc::~cPluginSc()
{
  delete status;
  cSoftCAM::ShutDown();
}

bool cPluginSc::CheckVersion(void)
{
  static char vv[] = VDRVERSION;
  int version, major, minor;
  if(sscanf(vv,"%d.%d.%d",&version,&major,&minor)==3) {
    if( version<MIN_VERS ||
       (version==MIN_VERS && major<MIN_MAJOR) ||
       (version==MIN_VERS && major==MIN_MAJOR && minor<MIN_MINOR)) {
      esyslog("ERROR: SC plugin needs at least VDR version %d.%d.%d",MIN_VERS,MIN_MAJOR,MIN_MINOR);
      fprintf(stderr,"SC plugin needs at least VDR version %d.%d.%d\n",MIN_VERS,MIN_MAJOR,MIN_MINOR);
      return false;
      }
    }
  else esyslog("ERROR: cannot parse VDR version string '%s'",vv);
  return true;
}

bool cPluginSc::Initialize(void)
{
  if(!CheckVersion()) return false;
  RegisterI18n(Phrases);
  if(!cSoftCAM::Startup(ConfigDirectory())) return false;
  status=new cScStatus;
  return true;
}

cMenuSetupPage *cPluginSc::SetupMenu(void)
{
  return new cMenuSetupSc(ConfigDirectory());
}

bool cPluginSc::SetupParse(const char *Name, const char *Value)
{
  if      (!strcasecmp(Name,"XorHack"))        ;
  else if (!strcasecmp(Name,"FirmwareHack"))   ;
  else if (!strcasecmp(Name,"LoggerActive"))   ScSetup.LoggerActive   = atoi(Value);
  else if (!strcasecmp(Name,"LoggerTimeout"))  ScSetup.LoggerTimeout  = atoi(Value);
  else if (!strcasecmp(Name,"UseFixedEMMPid")) ScSetup.UseFixedEMMPid = atoi(Value);
  else if (!strcasecmp(Name,"FixedEMMPid"))    ScSetup.FixedEMMPid    = atoi(Value);
  else return false;
  return true;
}

VDRPLUGINCREATOR(cPluginSc); // Don't touch this!
