/*
 * Softcam plugin to VDR (C++)
 *
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 */

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <limits.h>
#include <typeinfo>

#include <linux/dvb/ca.h>

#if defined(NAGRA_SUPPORT) | defined(CONAX_SUPPORT)
#include <openssl/bn.h>
#endif

#include <vdr/device.h>
#include <vdr/dvbdevice.h>
#include <vdr/videodir.h>
#include <vdr/channels.h>
#include <vdr/ci.h>

#include "common.h"
#include "softcam.h"
#include "logger.h"
#include "data.h"
#include "misc.h"
#include "setup.h"
#include "i18n.h"
#ifdef NAGRA_SUPPORT
#include "libnagra/nagra.h"
#endif

#ifdef BETACAM_SUPPORT
#include <assert.h>
#include "libbetacam/betacam.h"
#include "libbetacam/logger.h"
#endif // BETACAM_SUPPORT

// some sanity checks for applied patches
#ifdef HAVE_SOFTCSA
#if !defined(SOFTCSA_VERS) | SOFTCSA_VERS<5
#error You MUST use softcsa version 0.0.5 or greater. Next time read the README first.
#endif
#endif
#if !defined(VDR_IS_SC_PATCHED) | VDR_IS_SC_PATCHED<113
#error You MUST patch the VDR core with the supplied patch. Next time read the README first.
#endif


// -----------------------------------------------------------------------------

#define DEV_DVB_ADAPTER   "/dev/dvb/adapter"
#define DEV_DVB_CA        "ca"
#define DEV_DVB_DEMUX     "demux"

cEcmCache *ecmcache=0;
cPlainKeys *keys=0;

// -----------------------------------------------------------------------------

static const char *DvbName(const char *Name, int n)
{
  static char buffer[PATH_MAX];
  snprintf(buffer, sizeof(buffer), "%s%d/%s%d", DEV_DVB_ADAPTER, n, Name, 0);
  return buffer;
}

static int DvbOpen(const char *Name, int n, int Mode, bool ReportError=false)
{
  const char *FileName=DvbName(Name,n);
  int fd=open(FileName,Mode);
  if(fd<0 && ReportError) LOG_ERROR_STR(FileName);
  return fd;
}

// -- cECM ---------------------------------------------------------------------

class cECM : public cList<cECMInfo> {
private:
  int ecmDev;
  int prg, source, transponder;
  int cardNum;
  //
  cEcmData *curdat;
  cECMInfo *curecm;
  bool in_cache, abort;
  //
  void Clear(void);
  void ParseCAInfo(void);
  void ParseCADescriptor(unsigned char *data, int len);
  void AddECMInfo(int system, char *name, int ecm_pid, int id);
public:
  cECM(int CardNum, int EcmDev);
  ~cECM();
  cECMInfo *New(int PRG, int Source, int Transponder);
  cECMInfo *Jump(bool log=true);
  void Ok(void);
  void Abort(void) { abort=true; }
  };

cECM::cECM(int CardNum, int EcmDev)
{
  cardNum=CardNum; ecmDev=EcmDev;
}

cECM::~cECM()
{
  Clear();
}

void cECM::Clear(void)
{
  cListBase::Clear();
  curdat=0; curecm=0;
  in_cache=true; abort=false;
}

cECMInfo *cECM::New(int PRG, int Source, int Transponder)
{
  Clear();
  prg=PRG; source=Source; transponder=Transponder;
  return Jump();
}

cECMInfo *cECM::Jump(bool log)
{
  cECMInfo *e=0;
  if(in_cache) {
    cEcmData *dat=ecmcache->Search(prg,source,transponder,curdat);
    if(dat) { e=dat; curdat=dat; }
    else in_cache=false;
    }
  if(!e) {
    if(!curecm) {
      ParseCAInfo();
      e=First();
      }
    else e=Next(curecm);
    curecm=e;
    }
  if(e && log) d(printf("softcam %d: try system %s (%04x) id %04x with ecm %x%s\n",
                        cardNum,e->name,e->system,e->id,e->ecm_pid,(in_cache && curdat)?" (from cache)":""))
  return e;
}

void cECM::Ok(void)
{
  ecmcache->New(prg,source,transponder,in_cache?curdat:curecm);
  cSoftCAM::Save();
}

void cECM::ParseCAInfo(void)
{
  uchar buff[2048];
  int t=time_ms()+GETCA_TIMEOUT;
#ifdef DEBUG
  bool first=true;
#endif
  do {
    static const unsigned short casys[] = { 0xFFFF,0 };
    int len=cSIProcessor::GetCaDescriptors(source,transponder,prg,casys,sizeof(buff),buff);
    if(len>0) {
      d(printf("softcam %d: got CaDescriptors (len=%d time=%d)\n",cardNum,len,time_ms()-t+GETCA_TIMEOUT))
      int index=0;
      while(index<len) {
        int s=buff[index+1];
        if(buff[index]==0x09) ParseCADescriptor(&buff[index+2],s);
        s+=2; index+=s;
        }
      break;
      }
    else if(len<0) {
      d(printf("softcam %d: CA parse buffer overflow\n",cardNum))
      break;
      }
    else {
#ifdef DEBUG
      if(first) printf("softcam %d: waiting for CaDescriptors (timeout=%d)\n",cardNum,GETCA_TIMEOUT);
      first=false;
#endif
      usleep(200000);
      }
    } while (time_ms()<t && !abort);
}

void cECM::AddECMInfo(int system, char *name, int ecm_pid, int id)
{
  d(printf("softcam %d: found %s id %04x with ecm %x ",cardNum,name,id,ecm_pid))
  cECMInfo *e=First();
  while(e) {
    if(e->system==system && e->ecm_pid==ecm_pid && e->id==id) {
      d(printf("(already present)\n"))
      return;
      }
			e=Next(e);
    }
  d(printf("(new)\n"))
  if((e=new cECMInfo(name,system,ecm_pid,id))) Add(e);
  else d(printf("softcam %d: no memory for ecminfos\n",cardNum))
}

void cECM::ParseCADescriptor(unsigned char *data, int len)
{
  static char *seca_name     = "Seca";
  static char *viaccess_name = "Viaccess";
  static char *irdeto_name   = "Irdeto";
#ifdef NAGRA_SUPPORT
  static char *nagra_name    = "Nagra";
#endif
#ifdef CRYPTOWORKS_SUPPORT
  static char *cryptow_name  = "Cryptoworks";
#endif
#ifdef CONAX_SUPPORT
  static char *conax_name    = "Conax";
#endif

  int j;
  int ca_system=WORD(data,0,0xFFFF);

  switch(ca_system & MASK_CA_SYSTEM){
    case SECA_CA_SYSTEM:
      for(j=2 ; j<len ; j+=15) AddECMInfo(ca_system,seca_name,WORD(data,j,0x1FFF),WORD(data,j+2,0xFFFF));
      break;
    case VIACCESS_CA_SYSTEM:
      j=4;
      while(j<len) {
        if(data[j]==0x14) AddECMInfo(ca_system,viaccess_name,WORD(data,2,0x1FFF),(data[j+2]<<16)|(data[j+3]<<8)|(data[j+4]&0xF0));
        j+=2+data[j+1];
        }
      break;
    case BETA_CA_SYSTEM:
    case IRDETO_CA_SYSTEM:
      AddECMInfo(ca_system,irdeto_name,WORD(data,2,0x1FFF),0);
      break;
#ifdef NAGRA_SUPPORT
    case NAGRA_BEV_CA_SYSTEM:
    case NAGRA_CA_SYSTEM:
      AddECMInfo(ca_system,nagra_name,WORD(data,2,0x1FFF),0);
      break;
#endif
#ifdef CRYPTOWORKS_SUPPORT
    case CRYPTOW_CA_SYSTEM:
      AddECMInfo(ca_system,cryptow_name,WORD(data,2,0x1FFF),0);
      break;
#endif
#ifdef CONAX_SUPPORT
    case CONAX_CA_SYSTEM:
      AddECMInfo(ca_system,conax_name,WORD(data,2,0x1FFF),0);
      break;
#endif
    }
}

// -- cSoftCAM ----------------------------------------------------------------

#ifdef CRYPTOWORKS_SUPPORT
cMutex cSoftCAM::cryptow_lock;
#endif
struct CAMInfo cSoftCAM::camInfo[MAXDVBAPI] = {};
cMutex cSoftCAM::camLock;
int cSoftCAM::foundKeys=0;
int cSoftCAM::newKeys=0;
const char *cSoftCAM::cacheFile;

cSoftCAM::cSoftCAM(const cDevice *dev, int CardNum, int CamDev, int EcmDev, int EmmDev)
:cRunMode("softcam",CardNum)
{
  logger=0; loggerTime=0; softCSA=false;

  cardNum=CardNum;
  device=(cDevice *)dev;
  camDev=CamDev; ecmDev=EcmDev; emmDev=EmmDev;
#ifdef HAVE_SOFTCSA
  if(camDev<0) {
    softCSA=true;
    isyslog("Software decrypting activated on card %d",cardNum);
    }
#endif

  ecm=new cECM(cardNum,ecmDev);
  if(ecm) {
    prg=0; source=transponder=0; NoKey();
    Start();
    WaitRunMode(rmStartup,true);
    }
  else esyslog("ERROR: failed to create ECM class!");
}

cSoftCAM::~cSoftCAM()
{
  if(runMode!=rmKilled) {
    SetChannel(0);
    SetRunMode(rmStop);
    WaitRunMode(rmKilled,false);
    }
  delete logger;  
  delete ecm;
  if(camDev>0) close(camDev);
  if(ecmDev>0) close(ecmDev);
  if(emmDev>0) close(emmDev);
}

bool cSoftCAM::Startup(const char *cfgdir)
{
  if(!ecmcache) ecmcache=new cEcmCache();
  if(!ecmcache) {
    esyslog("ERROR: no memory for cache!");
    return false;
    }
  ecmcache->Load();

  if(!keys) keys=new cPlainKeys;
  if(!keys) {
    esyslog("ERROR: no memory for keys!");
    return false;
    }
  if(!keys->Load(cfgdir)) esyslog("ERROR: no keys loaded for softcam!");

  if(!cLogger::LoadCards(cfgdir)) return false;

  loaders.LoadCache(cfgdir);
  Save();
#ifdef NAGRA_SUPPORT
  SetRomDir(cfgdir);
#endif
  return true;
}

void cSoftCAM::Save(void)
{
  loaders.SaveCache();
}

void cSoftCAM::ShutDown(void)
{
  for(int n=0 ; n<MAXDVBAPI ; n++) {
    cSoftCAM *cam=cSoftCAM::camInfo[n].cam;
    if(cam) delete cam;
    }
#ifdef NAGRA_SUPPORT
  FreeRoms();
#endif
}

cSoftCAM *cSoftCAM::Init(const cDevice *dev)
{
  cSoftCAM *cam=0;
  int n=dev->CardIndex();

  if(n<MAXDVBAPI) {
    if(cSoftCAM::camInfo[n].shutUp) return 0;

#ifdef BETACAM_SUPPORT
		betacam_init(BETACAMPORTNAME);
#endif

    cSoftCAM::camLock.Lock();
  
    bool has_ca=false;
    for(int i=0; i<MAXCACAPS; i++) {
      int ca=Setup.CaCaps[n][i];
      if(ca==999) { has_ca=false; break; }
      else if(ca>MAXDVBAPI) has_ca=true;
      }
    if(has_ca) {
      if(typeid(*dev)==typeid(cDvbDevice)) {
        int ecmfd=DvbOpen(DEV_DVB_DEMUX, n, O_RDWR | O_NONBLOCK);
        int emmfd=DvbOpen(DEV_DVB_DEMUX, n, O_RDWR | O_NONBLOCK);
        int camfd=-1;
        cCiHandler *ci=((cDevice *)dev)->CiHandler();
        if(ci) { // get camDev from ciHandler
          camfd=ci->GetCaFd();
          d(printf("softcam %d: got camDev from ciHandler\n",n))
          }
        if(camfd<0) { // last resort, try to open myself
          camfd=DvbOpen(DEV_DVB_CA   , n, O_RDWR | O_NONBLOCK);
          d(printf("softcam %d: got camDev from myself (last resort)\n",n))
          }

        if(ecmfd>0
#ifndef HAVE_SOFTCSA
                   && camfd>0
#endif
                             ) {
          cam=new cSoftCAM(dev,n,camfd,ecmfd,emmfd);
          // ~cSoftCAM will close the devices
          }
        else esyslog("ERROR: can't open devices for softcam on card %d",n);
        }
      else esyslog("ERROR: softcam doesn't support card %d!",n);
      }

    if(!cam) cSoftCAM::camInfo[n].shutUp=true;
    cSoftCAM::camLock.Unlock();
    }
  return cam;
}

cSoftCAM *cSoftCAM::FindCAM(const cDevice *dev)
{
  int n=dev->CardIndex();
  cSoftCAM::camLock.Lock();
  cSoftCAM *cam=cSoftCAM::camInfo[n].cam;
  if(n<MAXDVBAPI && cam) {
    if(cam->device!=dev) {
      cam=0;
      esyslog("ERROR: internal error, cam->device!=device in FindCAM()!");
      }
    }
  cSoftCAM::camLock.Unlock();
  return cam;
}

const char *cSoftCAM::CurrKeyStr(int CardNum)
{
  cSoftCAM::camLock.Lock();
  cSoftCAM *cam=camInfo[CardNum].cam;
  char *str=cam ? cam->currentKeyStr : 0;
  cSoftCAM::camLock.Unlock();
  return str;
}

bool cSoftCAM::Active(void)
{
  bool res=false;
  cSoftCAM::camLock.Lock();
  for(int n=0 ; n<MAXDVBAPI ; n++) {
    cSoftCAM *cam=cSoftCAM::camInfo[n].cam;
    if(cam && cam->CurrentPrg()) { res=true; break; }
    }
  cSoftCAM::camLock.Unlock();
  return res;
}

void cSoftCAM::SetChannel(int Prg, int Source, int Transponder)
{
  d(printf("softcam %d: setting new SID %d, source %x, transponder %x ",cardNum,Prg,Source,Transponder))
  if((Prg!=prg || Source!=source || Transponder!=transponder) &&
     (Prg>0 || !device->Receiving(true))) {
    d(printf("(send)\n"))
    newprg=Prg; newSource=Source; newTransponder=Transponder;
    ecm->Abort();
    SetRunMode(rmChange);
    WaitRunMode(rmChange,true,10*1000);
    if(runMode==rmChange && runMode!=rmKilled) d(printf("softcam %d: timeout/error setting program number\n",cardNum))
    }
  else d(printf("(ignored)\n"))
}

void cSoftCAM::WriteCW(int index, bool resend)
{
  ca_descr_t ca_descr;
  int i;

  for(i=0; i<16; i+=4) cw[i+3]=cw[i]+cw[i+1]+cw[i+2];

#ifdef DEBUG_KEYS
  char str1[20], str2[20];
  printf("softcam %d: even cw %s odd cw %s\n",cardNum,KeyStr(str1,&cw[0]),KeyStr(str2,&cw[8]));
#endif

  ca_descr.index=index;
  if(resend || memcmp(&cw[0],lastEvenCW,8)) {
    memcpy(lastEvenCW,&cw[0],8);
    ca_descr.parity=0;
    for(i=7; i>=0; i--) ca_descr.cw[i]=cw[i];
    if(!softCSA) { CHECK(ioctl(camDev,CA_SET_DESCR,&ca_descr)); }
#ifdef HAVE_SOFTCSA
    else device->SetCADescr(&ca_descr);
#endif
    DTURN('E')
    }

  if(resend || memcmp(&cw[8],lastOddCW,8)) {
    memcpy(lastOddCW,&cw[8],8);
    ca_descr.parity=1;
    for(i=7; i>=0; i--) ca_descr.cw[i]=cw[i+8];
    if(!softCSA) { CHECK(ioctl(camDev,CA_SET_DESCR,&ca_descr)); }
#ifdef HAVE_SOFTCSA
    else device->SetCADescr(&ca_descr);
#endif
    DTURN('O')
    }
}

void cSoftCAM::Action()
{
  dsyslog("softcam thread started on card %d (pid=%d)", cardNum, getpid());
  camInfo[cardNum].pid=getpid(); camInfo[cardNum].cam=this;
  SetRunMode(rmRun);

  while(runMode!=rmStop) {
    WaitRunMode(rmChange,false);
#ifdef DEBUG_TURN
    static const char turn[] = { '/','-','\\','|','X' };
    int turns=0, lastCwTime=0;
    printf("softcam %d: loop\n",cardNum);
#endif

    sync=false; currentKey=0; dolog=LOG_COUNT;
    if(runMode==rmChange) {
      prg=newprg; source=newSource; transponder=newTransponder;
      SetRunMode(rmRun);
      }
    if(prg<0) {
      prg=0;                // don't deadlock with the following channel switch
      cChannel *Channel;    // switch the channel to FTA
      int i=1;
      do {
        Channel=Channels.GetByNumber(i++);
        } while(Channel && Channel->Ca()>=MAXDEVICES); 
      if(Channel) {
        device->SwitchChannel(Channel,true);
        d(printf("softcam %d: switched to FTA\n",cardNum))
        }
      }
    if(!prg) continue;

#ifdef DEBUG
    int startsync=time_ms();
#endif
    cECMInfo *e=ecm->New(prg,source,transponder);
    while(!e && runMode==rmRun) {
      d(printf("softcam %d: no encryption system found\n",cardNum))
      WaitRunMode(rmChange,false,CA_SEARCH_TIME*1000);
      if(runMode==rmRun) e=ecm->Jump(true);
      }
    if(!e) continue;

    SetFilter(ecmDev,e->ecm_pid,false,0x80,0xfe,0);
    SetLogger(e->system,e->id);
    Init();

    unsigned char parity=0xFF;
    int count=0;
    int resendTime=0, receive=0;
    while(runMode==rmRun) {
#ifdef DEBUG_TURN
      printf("\b%c",turn[turns]); fflush(stdout);
      turns=(turns+1)%sizeof(turn);
#endif
      unsigned char buffer[4096];
      int n=ReadPaket(ecmDev,buffer,sizeof(buffer),sync?SYNC_TIMEOUT:NOSYNC_TIMEOUT);
      if(n>0) {
#ifdef ECM_DUMP
        Dump(ECM_DUMP,buffer,n);
#endif
        bool resend=false;
        if(resendTime && time_ms()>=resendTime) {
          DTURN('T')
          resend=true; resendTime=0;
          }

        if(buffer[0]==0x80 || buffer[0]==0x81) {
          if(PreCheck(e,buffer)) {
            if(resend || parity!=(buffer[0]&1)) {
              if(Process(e,buffer)) {
                WriteCW(0,resend || !sync);
                if(!resend) {
#ifdef DEBUG_TURN
                  int i=lastCwTime;
                  lastCwTime=time_ms();
                  if(i) { printf("\b[%d] ",lastCwTime-i); fflush(stdout); }
#endif
                  }
                if(!sync) {
                  d(printf("softcam %d: correct key found\n",cardNum))
                  ecm->Ok(); sync=true;
#ifdef DEBUG
                  d(printf("softcam %d: sync time %d ms\n",cardNum,time_ms()-startsync))
                  //system("echo >/dev/tty0 -e \"\\a\"");
#endif
                  if(!softCSA) {
                    resendTime=time_ms()+CW_REPEAT_TIME;
                    if(device->Receiving(true)) receive=1;
                    }
                  if(!receive) SetLogger(e->system,e->id,true);
                  }
                parity=buffer[0]&1;
                DTURN('+')
                }
              else {
                parity=0xFF;
                if(!sync && ++count>=ECM_TRY_COUNT) n=-1;
                if(dolog) DTURN('>')
                }
              }
            }
          else {
            sync=false; count=0; parity=0xFF;
            }
          }
        }
      if(!sync && n<=0) { // try next ecm-pid if an error occured and we haven't got a valid key before
        StopFilter(ecmDev); NoKey(); count=0;
#ifdef DEBUG
#define JUMPARG (dolog!=0)
#else
#define JUMPARG
#endif
        e=ecm->Jump(JUMPARG);
        if(!e) {
#ifdef DEBUG
          if(dolog) {
            dolog--;
            if(!dolog) printf("softcam %d: stopping console log until valid key is found\n",cardNum);
            }
#endif
          e=ecm->Jump(JUMPARG);
          if(!e) break;
          }
        SetFilter(ecmDev,e->ecm_pid,false,0x80,0xfe,0);
        SetLogger(e->system,e->id);
        }
      if(sync && receive>=0 && device->Receiving(true)) {
        receive=-1;
        if(!softCSA && ScSetup.LoggerActive>1) {
          d(printf("softcam %d: receive mode detected\n",cardNum))
          SetLogger(0,0);  // stop logging during recordings/transfermode
          }
        }
      if(receive<0 && !device->Receiving(true)) {
        d(printf("softcam %d: receive mode ended\n",cardNum))
        SetLogger(e->system,e->id,true);
        receive=0;
        }
      }
    StopFilter(ecmDev); NoKey();
    SetLogger(0,0);
    DeInit();
    }

  SetRunMode(rmKilled);
  camLock.Lock();
  camInfo[cardNum].pid=0; camInfo[cardNum].cam=0;
  camLock.Unlock();
  dsyslog("softcam thread ended on card %d (pid=%d)", cardNum, getpid());
}

void cSoftCAM::SetLogger(int caSys, int Id, bool force)
{
  if(!logger && ScSetup.LoggerActive && caSys && emmDev>=0) {
    logger=new cLogger(cardNum,emmDev);
    }
  if(!ScSetup.LoggerActive) {
    delete logger; logger=0;
    }
  if(logger && (force || !caSys || time(0)>=loggerTime)) {
    if(caSys) loggerTime=time(0)+MIN_CHANGE_TIME; else loggerTime=0;
    logger->SetCaSys(caSys,Id);
    }
}

// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------

void cSoftCAM::Init(void)
{
  i_flag=false;
  memset(i_chid,0,sizeof(i_chid));
#ifdef NAGRA_SUPPORT
  ClearNagra();
#endif
}

void cSoftCAM::DeInit(void)
{
}

bool cSoftCAM::PreCheck(cECMInfo *ecm, unsigned char *buffer)
{
  switch(ecm->system & MASK_CA_SYSTEM) {
    case IRDETO_CA_SYSTEM:
      //
      // search for fake channel id's on Stream
      //
      {
      const int max=buffer[5];
      if(max>=I_CHID_SIZE) d(printf("softcam %d: max chid > %d\n",cardNum,I_CHID_SIZE))
      if(max>0 && max<I_CHID_SIZE) {
        const int num=buffer[4];
        if(num==0) i_flag=true;
        if(num>=I_CHID_SIZE) d(printf("softcam %d: curr chid > %d\n",cardNum,I_CHID_SIZE))
        if(i_flag && num<I_CHID_SIZE) i_chid[num]=WORD(buffer,6,0xFFFF);
        if(num==max) {
          for(int i=0 ; i<max ; i++) {
            if(i_chid[i]==0) {
              d(printf("softcam %d: zero channel id\n",cardNum))
              i_flag=false;
              return false;
              }
            for(int j=i+1 ; j<=max ; j++) {
              if(i_chid[i]==i_chid[j]) {
                d(printf("softcam %d: fake channel id %04x\n",cardNum,i_chid[i]))
                i_flag=false;
                return false;
                }
              }
            }
          }
        }
      return true;
      }
    default:
      return true;
    }
}

bool cSoftCAM::Process(cECMInfo *ecm, unsigned char *buffer)
{
  bool r=false;
  firstKey=-1; cw=cwMem;
  switch(ecm->system & MASK_CA_SYSTEM) {
    case SECA_CA_SYSTEM:
      return Seca(&buffer[8],buffer[2]-5,ecm->id,buffer[7] & 0x0f);
    case VIACCESS_CA_SYSTEM:
      return Viaccess(&buffer[4],buffer[2]-1,ecm->id);
#ifndef BETACAM_SUPPORT
    case IRDETO_CA_SYSTEM:
    case BETA_CA_SYSTEM:
      return Irdeto(&buffer[6]);
#else  // BETACAM_SUPPORT
    case BETA_CA_SYSTEM:
 			return betacam_ecm(buffer, cwMem);
    case IRDETO_CA_SYSTEM:
      return Irdeto(&buffer[6]);
#endif // BETACAM_SUPPORT
#ifdef NAGRA_SUPPORT
    case NAGRA_BEV_CA_SYSTEM:
    case NAGRA_CA_SYSTEM:
      return Nagra(&buffer[5],buffer[4]);
      break;
#endif
#ifdef CRYPTOWORKS_SUPPORT
     case CRYPTOW_CA_SYSTEM:
      cryptow_lock.Lock(); // cryptoworks can only be use by one card a time !!
      r=(cryptoworks_decode(cwMem,buffer,ISO_TTY)==0);
      cryptow_lock.Unlock();
      break;
#endif
#ifdef CONAX_SUPPORT
    case CONAX_CA_SYSTEM:
      return Conax(&buffer[7],buffer[4]-2,buffer[6]);
#endif
    }
  return r;
}

cPlainKey *cSoftCAM::FastKey(int type, int id, int keynr)
{
  if(currentKey && currentKey->type==type && currentKey->id==id && currentKey->keynr==keynr) {
    if(firstKey<=0) {
      if(firstKey<0) { firstKey=0; return currentKey; }
      else currentKey=0;
      }
    }
  else currentKey=0;

  currentKey=keys->FindKey(type,id,keynr,currentKey);
  firstKey=1;
  return currentKey;
}

void cSoftCAM::NoKey(void)
{
  strcpy(currentKeyStr,tr("None"));
  lastType=0;
}

void cSoftCAM::KeyOK(cPlainKey *pk)
{
  if(pk->type!=lastType || pk->id!=lastId || pk->keynr!=lastKeynr) {
    char *str=pk->Print();
    lastType=pk->type; lastId=pk->id; lastKeynr=pk->keynr;
    snprintf(currentKeyStr,sizeof(currentKeyStr),"%c %.2x %02x %s\n",lastType,lastId,lastKeynr,str);
#ifdef DEBUG
    dolog=LOG_COUNT;
    printf("softcam %d: using key %c %.2x %02x %s\n",cardNum,lastType,lastId,lastKeynr,str);
#endif
    free(str);
    }
}

void cSoftCAM::KeyFail(int type, int id, int keynr)
{
#ifdef DEBUG
  if(type!=lastType || id!=lastId || keynr!=lastKeynr) {
    lastType=type; lastId=id; lastKeynr=keynr;
    if(dolog)
      printf("softcam %d: no key found for %c %.2x %02x\n",cardNum,lastType,lastId,lastKeynr);
    }
#endif
}

bool cSoftCAM::Irdeto(unsigned char *source)
{
  unsigned char *data=0;
  int date=-1, keynr=0;
  int length=source[5]+6-5; // 6 header bytes - 5 signature bytes
  for(int index=6 ; index<length && (data==0 || date==-1) ;) {
    int param=source[index++];
    int len  =source[index++] & 0x3f;

    switch(param) {
      case 0x78:
        keynr = source[index];
        data = &source[index+2];
        break;
      case 0x00:
      case 0x40:
        date = (source[index]<<8) | source[index+1];
        break;
      }
    index+=len;
    }
  if(data==0 || date==-1) return false;

  cPlainKey *pk;
  while((pk=FastKey('I',source[2],keynr))) {
    unsigned char save[16];
    unsigned char key[8];
    memcpy(save,data,16); // save the encrypted data
    pk->Get(key);
    sessionKeyCrypt(&data[0],key,date);
    sessionKeyCrypt(&data[8],key,date);
    if(signatureCheck(source,length,key,date,&source[length],8)) {
      cw=&data[0];
      KeyOK(pk);
      return true;
      }
    memcpy(data,save,16); // put back the encrypted data if it didn't works
    }
  d(KeyFail('I',source[2],keynr))
  return false;
}

bool cSoftCAM::Viaccess(unsigned char *source, int length, int id)
{
  static unsigned char table0[] = {
    0x4F,0xB4,0xFC,0x9B,0x4A,0x7F,0x44,0xFB,0x05,0xFF,0xBD,0xBB,0x16,0x2D,0x6C,0xC8,
    0xD8,0x96,0xF9,0xFE,0x3F,0xFF,0x36,0x24,0xB6,0xBF,0x49,0xC9,0x2D,0x36,0x5E,0xD0,
    0x1F,0x09,0x7E,0xA9,0x7F,0xFF,0x64,0xB6,0x5B,0x7E,0xF8,0xFC,0x6E,0x3F,0x7F,0xBF,
    0xDD,0x36,0x12,0xE9,0x05,0xFE,0xB4,0x6C,0x6F,0xFE,0x7E,0xC8,0x25,0x90,0x6D,0x90
    };
  static unsigned char table1[] = {
    0x7E,0x6D,0x7E,0x12,0x76,0xFD,0x2F,0xFE,0x6D,0xFE,0xDA,0x3F,0xDA,0x6D,0xBD,0x97,
    0xD0,0x6D,0xD8,0x9F,0x69,0xFD,0xB6,0x37,0xFE,0x7F,0x36,0x92,0xBD,0x52,0x16,0xDF,
    0xFC,0x96,0xFF,0x92,0xFD,0x6D,0x7F,0xB5,0xFB,0x4C,0xB6,0xB7,0x7E,0xD9,0xFE,0x9B,
    0xFD,0xF4,0x6D,0x9B,0xB9,0x36,0xBF,0x7F,0xD2,0x2D,0xDF,0xB7,0xD9,0xFE,0x69,0xBF
    };

  /* maybe get this nano first */
  if(source[0]==0xd2) {
    source+=3;
    length-=3;
    unsigned char *table=source[0]==0xdf ? table0 : table1;
    for(int i=0 ; i<length ; i++) source[i]^=table[i&63];
    }

  /* next should get 0x90 nano (length 3) containing id and key */
  int keynr=0;
  if(source[0]==0x90 && source[1]==0x03) {
    keynr=source[4] & 0x0f;
    source+=5;
    length-=5;

    cPlainKey *pk=FastKey('V',id,keynr);
    if(pk) {
      unsigned char key[8];
      pk->Get(key);
      if(decrypt(key,source,length,&cw[0],&cw[8])) {
        KeyOK(pk);
        return true;
        }
      }
    }
  d(KeyFail('V',id,keynr))
  return false;
}

bool cSoftCAM::Seca(unsigned char *source, int length, int id, int keynr)
{
  cPlainKey *pk;
  while((pk=FastKey('S',id,keynr))) {
    unsigned char keys[16];
    pk->Get(keys);
    memcpy(&keys[8],&keys[0],8); // duplicate key
    unsigned char signature[8];
    memset(signature,0,sizeof(signature));
    for(int i=0 ; i<length-8 ; i+=8) {
      for(int j=0 ; j<8 && i+j<length-8; j++) signature[j]^=source[i+j];
      encrypt_seca(keys,signature);
      }

    unsigned char *data=0;
    for(int i=0 ; i<length ; ) {
      int param=source[i++];
      int extra=(param >> 4) & 0x0f;
      switch(extra) {
        case 0x0d: extra=0x10; break;
        case 0x0e: extra=0x18; break;
        case 0x0f: extra=0x20; break;
        }

      switch(param) {
        case 0xd1:
          data = &source[i];
          break;
        case 0x82:
          if(memcmp(&source[i],signature,8)) { data=0; i=length; }
          break;
        }
      i+=extra;
      }
    if(data) {
      decrypt_seca(&keys[0],&data[0]);
      decrypt_seca(&keys[0],&data[8]);
      cw=&data[0];
      KeyOK(pk);
      return true;
      }
    }
  d(KeyFail('S',id,keynr))
  return false;
}

#ifdef CONAX_SUPPORT

void cSoftCAM::ConaxParseECM(unsigned char *buf, int len)
{
  int i=0;
  while(i<len) {
    int nano=buf[i++];
    int nanolen=buf[i++];
    
    switch(nano) {
      case 0x80:
      case 0x82:
        ConaxParseECM(buf+i+3,nanolen-3); break;
      case 0x81:
        ConaxParseECM(buf+i+2,nanolen-2); break;
      case 0x30:
	memcpy(cwMem  ,buf+i+10,8);
	memcpy(cwMem+8,buf+i+2 ,8);
        break;
      }
    i+=nanolen;
    }
}

bool cSoftCAM::Conax(unsigned char *source, int length, int keynr)
{
  static const unsigned char hash[] = { 0x05,0x00,0x05 };

  cPlainKey *pk;
  if(!(pk=FastKey('C','M',keynr))) return false;
  BIGNUM *mod=BN_new();
  pk->Get(mod);

  if(!(pk=FastKey('C','E',keynr))) return false;
  BIGNUM *exp=BN_new();
  pk->Get(exp);

  BN_CTX *temp=BN_CTX_new();
  BIGNUM *r=BN_new();
  unsigned char tempc[256*3];
  int i=length;

  while(i>0) {
    int index=i-64; if(index<0) index=0;

    BIGNUM *ecm=BN_bin2bn(source+index,64,0);
    BN_mod_exp(r,ecm,exp,mod,temp);

    int cnt=BN_bn2bin(r,tempc);
    if(cnt<64) memset(source+index,0,64-cnt);
    memcpy(source+index+64-cnt,tempc,cnt);

    i-=(i%60 && index) ? i%60 : 64;
    BN_free(ecm);
    }

  BN_free(r); BN_free(mod); BN_free(exp);
  BN_CTX_free(temp);

  if(memcmp(hash,source+2,3) || memcmp(source+5,source+length-5,3))
    return false; // signature incorrect

  ConaxParseECM(source+10,length-10);
  return true;
}

#endif //CONAX_SUPPORT

#ifdef NAGRA_SUPPORT

void cSoftCAM::WriteNagraTable(unsigned char *from, int off)
{
  if(off+16<256) memcpy(Nagra_MECMTable+off,from,16);
  else {
    int l=256-off;
    memcpy(Nagra_MECMTable+off,from,l);
    memcpy(Nagra_MECMTable,from+l,16-l);
    }
}

void cSoftCAM::ClearNagra(void)
{
  memset(Nagra_MECMTable,0,sizeof(Nagra_MECMTable));
}

bool cSoftCAM::Nagra(unsigned char *data, int len)
{
  const int id=(data[0]&0xFE)*256 + data[1];
  const int keynr=(data[4]&0x10)>>4;

  unsigned char sessionKey[8], verifyKey[8];
  bool hasvkey=false;
  cPlainKey *pk;
  if((pk=FastKey('N',id,0x80))) {
    pk->Get(verifyKey);
    hasvkey=true;
    }
  if(!(pk=FastKey('N',id,keynr))) return false;
  pk->Get(sessionKey);

  unsigned char decrypted[128];
  unsigned char const *pdata=data+5+8;
  const int plen=(len-5-8)/8;
  for(int i=0 ; i<plen ; i++)
    Nagra_Decrypt(pdata+i*8,sessionKey,decrypted+i*8);

  if((decrypted[0]!=0x10 && decrypted[0]!=0x11 && decrypted[0]!=0x12) ||
     (hasvkey && !Nagra_SigCheck(decrypted,data+5,verifyKey,plen)))
    return false;

  unsigned char *pmecm=data+data[3]+4;
  const int plenmecm=pmecm[1]-4;
  if(len>data[3]+4 && *pmecm==0x20) {
    const int mecmprovid=pmecm[2]*256+pmecm[3];

    BIGNUM *e1 = BN_new();
    if(!(pk=FastKey('N',mecmprovid,MBC('E','1')))) return false;
    pk->Get(e1);
    BIGNUM *n1 = BN_new();
    if(!(pk=FastKey('N',mecmprovid,MBC('N','1')))) return false;
    pk->Get(n1);
    BIGNUM *n2 = BN_new();
    if(!(pk=FastKey('N',mecmprovid,MBC('N','2')))) return false;
    pk->Get(n2);

    if (!Nagra_RSADecrypt_ECM(&pmecm[6],pmecm[4],plenmecm,e1,n1,n2))
        return false;

    unsigned char mecmdata[128];
    memcpy(mecmdata, &pmecm[6], plenmecm);

    if(mecmdata[0]==0x2F && mecmdata[1]==pmecm[2] && mecmdata[2]==pmecm[3])
      WriteNagraTable(mecmdata+4,mecmdata[3]*2);

    BN_free(e1); BN_free(n1); BN_free(n2);
    }

  if(decrypted[1]<0x80) {
    int d=decrypted[ 1]*2;
    for(int i=0 ; i<8 ; i++) decrypted[i+2] ^=Nagra_MECMTable[(d+i)&0xFF];
    }
  if(decrypted[10]<0x80) {
    int d=decrypted[10]*2;
    for(int i=0 ; i<8 ; i++) decrypted[i+11]^=Nagra_MECMTable[(d+i)&0xFF];
    }

  memcpy(cwMem  ,decrypted+11,8);
  memcpy(cwMem+8,decrypted+2 ,8);
  return true;
}

#endif //NAGRA_SUPPORT
